﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Kocic.RepositoryManagement.LinqToSql;
using Sample.LinqToSql.Model;
using Kocic.RepositoryManagement;
using System.Configuration;

namespace Sample.Run
{
    public static class LinqToSqlSamples
    {
        public static void Run()
        {
            #region Connection Strings
            String DATA_CONNECTION_STRING = ConfigurationManager.ConnectionStrings["SampleDataModelClasses"].ToString();
            #endregion

            #region Cleanup
            Console.WriteLine("LINQ to SQL Samples ---------------------------\n");

            ////Remove any existing data
            using (new LinqToSqlExecutionContext<SampleDataModelClassesDataContext>(DATA_CONNECTION_STRING))
            {
                LinqToSqlRepositoryManager<Sample.LinqToSql.Model.Person, SampleDataModelClassesDataContext>.DeleteDataObjects(p => p);
            }
            
            #endregion

            #region Example #1 : Basic Context Invocation and Person Creation

            Console.WriteLine("Example: #1 ------------------------------------------");

            SampleDataModelClassesDataContext context = new SampleDataModelClassesDataContext(DATA_CONNECTION_STRING);

            //Create instance of person manager and delete it right after. PersonManager disposes the context immediatly after existing this code block.
            using (Sample.Run.LinqToSql.PersonRepositoryManager personManager = new Sample.Run.LinqToSql.PersonRepositoryManager(context))
            {
                //Event subscribing
                personManager.Created += manager_LinqToSqlCreated;
                personManager.Deleted += manager_LinqToSqlDeleted;

                //Create new person
                Sample.LinqToSql.Model.Person p1 = new Sample.LinqToSql.Model.Person();
                p1.FirstName = "Nikola";
                p1.MiddleName = "Vlajko";
                p1.LastName = "Kocic";
                p1.Age = 29;

                personManager.Save(p1, SaveMethod.Create);

                //Delete person
                personManager.Delete(p1);
            }

            #endregion

            #region Example #2 : Using the LinqToSqlExecutionContext
            Console.WriteLine("\n\nExample: #2 ------------------------------------------");

            using (new LinqToSqlExecutionContext<SampleDataModelClassesDataContext>(DATA_CONNECTION_STRING))
            {
                //Create instance of person manager and delete it right after. PersonManager uses the single instance of the context as specified in the EntityExecutionContext
                using (Sample.Run.LinqToSql.PersonRepositoryManager personManager = new Sample.Run.LinqToSql.PersonRepositoryManager())
                {
                    //Event subscribing
                    personManager.Created += manager_LinqToSqlCreated;
                    personManager.Deleted += manager_LinqToSqlDeleted;
                    personManager.Updated += manager_LinqToSqlUpdated;

                    //Create new person
                    Sample.LinqToSql.Model.Person p1 = new Sample.LinqToSql.Model.Person();
                    p1.FirstName = "Nikola";
                    p1.MiddleName = "Vlajko";
                    p1.LastName = "Kocic";

                    //Create person - Explicit call. Need to specify save method because the data context tracks the attached/detached internally
                    personManager.Save(p1, SaveMethod.Create);

                    p1.Age = 29;

                    //Update person - Implicit call.
                    personManager.Save(p1, SaveMethod.Update);

                    //Delete person
                    personManager.Delete(p1);
                }
            }

            #endregion

            #region Example #3 : Deferring Operations
            Console.WriteLine("\n\nExample: #3 ------------------------------------------");

            using (new LinqToSqlExecutionContext<SampleDataModelClassesDataContext>(DATA_CONNECTION_STRING))
            {
                Person p;
                Person p2;

                using (Sample.Run.LinqToSql.PersonRepositoryManager createManager = new Sample.Run.LinqToSql.PersonRepositoryManager())
                {
                    //Defer all data operations until exit of code block
                    using (new LinqToSqlContextDeferrer<SampleDataModelClassesDataContext>())
                    {
                        //Subscribe to creating event instead of Created since operation is being deferred
                        createManager.Creating += manager_LinqToSqlCreating;
                        p = new Sample.LinqToSql.Model.Person();
                        p.FirstName = "Nikola";
                        p.LastName = "Kocic";

                        createManager.Save(p, SaveMethod.Create);

                        p2 = new Sample.LinqToSql.Model.Person();
                        p2.FirstName = "John";
                        p2.LastName = "Doe";

                        createManager.Save(p2, SaveMethod.Create);

                        Console.WriteLine("Before Deferral: Person count in database: " + createManager.RetrieveMultiple().ToList().Count.ToString());
                    }
                    Console.WriteLine("After Deferral: Person count in database: " + createManager.RetrieveMultiple().ToList().Count.ToString());
                }

                using (Sample.Run.LinqToSql.PersonRepositoryManager deleteManager = new Sample.Run.LinqToSql.PersonRepositoryManager())
                {
                    //Subscribe to deleting event instead of Deleted since operation is being deferred
                    deleteManager.Deleting += manager_LinqToSqlDeleting;

                    deleteManager.Delete(p);
                    deleteManager.Delete(p2);

                    Console.WriteLine("After Deferral: Person count in database: " + deleteManager.RetrieveMultiple().ToList().Count.ToString());
                }
            }

            #endregion

            #region Example #4 : Creating and Deleting a Person Entity Across Two Different Contexts
            Console.WriteLine("\n\nExample: #4 ------------------------------------------");

            Sample.LinqToSql.Model.Person person4;
            using (new LinqToSqlExecutionContext<SampleDataModelClassesDataContext>(DATA_CONNECTION_STRING))
            {
                using (Sample.Run.LinqToSql.PersonRepositoryManager manager = new Sample.Run.LinqToSql.PersonRepositoryManager())
                {
                    Sample.LinqToSql.Model.Person temp;
                    manager.Created += manager_LinqToSqlCreated;

                    temp = new Sample.LinqToSql.Model.Person();
                    temp.FirstName = "Nikola";
                    temp.MiddleName = "Vlajko";
                    temp.LastName = "Kocic";

                    manager.Save(temp, SaveMethod.Create);  //Explicitly specify save method

                    person4 = manager.RetrieveSingle("Nikola", "Kocic");
                }
            }

            //At this point person4 instance doesn't belong to the same context below and the ObjectStateManager will not be able to find it.
            //Because of this, the PreDelete operation is implemented in the PersonManager class which gets the data object from the current context
            using (Sample.Run.LinqToSql.PersonRepositoryManager manager = new Sample.Run.LinqToSql.PersonRepositoryManager(new SampleDataModelClassesDataContext(DATA_CONNECTION_STRING)))
            {
                manager.Deleted += manager_LinqToSqlDeleted;
                manager.Delete(person4);

                Console.WriteLine("Person count in database: " + manager.RetrieveMultiple().ToList().Count.ToString());
            }

            #endregion

            #region Example #5 : Static Operations

            Console.WriteLine("\n\nExample: #5 ------------------------------------------");

            Sample.LinqToSql.Model.Person person5 = new Sample.LinqToSql.Model.Person();
            person5.FirstName = "Nikola";
            person5.MiddleName = "Vlajko";
            person5.LastName = "Kocic";

            using (new LinqToSqlExecutionContext<SampleDataModelClassesDataContext>(DATA_CONNECTION_STRING))
            {
                //Save data - Static Method
                LinqToSqlRepositoryManager<Sample.LinqToSql.Model.Person, SampleDataModelClassesDataContext>.SaveDataObject(person5, SaveMethod.Create);

                Console.WriteLine("Person count in database after static save: " + LinqToSqlRepositoryManager<Sample.LinqToSql.Model.Person, SampleDataModelClassesDataContext>.Retrieve(p => p).ToList().Count.ToString());

                //Delete saved person
                LinqToSqlRepositoryManager<Sample.LinqToSql.Model.Person, SampleDataModelClassesDataContext>.DeleteDataObject(person5);

                Console.WriteLine("Person count in database after static delete:" + LinqToSqlRepositoryManager<Sample.LinqToSql.Model.Person, SampleDataModelClassesDataContext>.Retrieve(p => p).ToList().Count.ToString());
            }

            #endregion
        }

        #region LINQ to SQL Event Methods
        private static void manager_LinqToSqlCreated(object sender, DataObjectEventArgs<Sample.LinqToSql.Model.Person> e)
        {
            Console.WriteLine();
            Console.WriteLine(MessageHelper.BuildPersonDataMessage("Created", e.DataObject.FirstName, e.DataObject.MiddleName, e.DataObject.LastName, e.DataObject.Age.ToString()));
        }

        private static void manager_LinqToSqlDeleted(object sender, DataObjectEventArgs<Sample.LinqToSql.Model.Person> e)
        {
            Console.WriteLine();
            Console.WriteLine(MessageHelper.BuildPersonDataMessage("Deleted", e.DataObject.FirstName, e.DataObject.MiddleName, e.DataObject.LastName, e.DataObject.Age.ToString()));
        }

        private static void manager_LinqToSqlUpdated(object sender, DataObjectEventArgs<Sample.LinqToSql.Model.Person> e)
        {
            Console.WriteLine();
            Console.WriteLine(MessageHelper.BuildPersonDataMessage("Updated", e.DataObject.FirstName, e.DataObject.MiddleName, e.DataObject.LastName, e.DataObject.Age.ToString()));
        }

        private static void manager_LinqToSqlCreating(object sender, DataObjectEventArgs<Sample.LinqToSql.Model.Person> e)
        {
            Console.WriteLine();
            Console.WriteLine(MessageHelper.BuildPersonDataMessage("Creating", e.DataObject.FirstName, e.DataObject.MiddleName, e.DataObject.LastName, e.DataObject.Age.ToString()));
        }

        private static void manager_LinqToSqlDeleting(object sender, DataObjectEventArgs<Sample.LinqToSql.Model.Person> e)
        {
            Console.WriteLine();
            Console.WriteLine(MessageHelper.BuildPersonDataMessage("Deleting", e.DataObject.FirstName, e.DataObject.MiddleName, e.DataObject.LastName, e.DataObject.Age.ToString()));
        }

        private static void manager_LinqToSqlUpdating(object sender, DataObjectEventArgs<Sample.LinqToSql.Model.Person> e)
        {
            Console.WriteLine();
            Console.WriteLine(MessageHelper.BuildPersonDataMessage("Updating", e.DataObject.FirstName, e.DataObject.MiddleName, e.DataObject.LastName, e.DataObject.Age.ToString()));
        }

        #endregion
    }
}
